<?php

// phpcs:disable
use CRM_Loginsecurity_ExtensionUtil as E;
// phpcs:enable

class CRM_Loginsecurity_BAO_LoginSecurityDevice extends CRM_Loginsecurity_DAO_LoginSecurityDevice {

  /**
   * Create a new LoginSecurityDevice based on array-data
   *
   * @param array $params key-value pairs
   * @return CRM_Loginsecurity_DAO_LoginSecurityDevice|NULL
   *
  public static function create($params) {
    $className = 'CRM_Loginsecurity_DAO_LoginSecurityDevice';
    $entityName = 'LoginSecurityDevice';
    $hook = empty($params['id']) ? 'create' : 'edit';

    CRM_Utils_Hook::pre($hook, $entityName, CRM_Utils_Array::value('id', $params), $params);
    $instance = new $className();
    $instance->copyValues($params);
    $instance->save();
    CRM_Utils_Hook::post($hook, $entityName, $instance->id, $instance);

    return $instance;
  } */

  /**
   * Updates the current session (if it has been more than 15 minutes), logging the IP and fingerprint.
   *
   * Returns the contact_id if it is a new/updated session (used to determine whether to load the fingerprinting JS).
   */
  public static function logCurrentSession() {
    // Check if the device is already registered
    $session_id = session_id();
    $ip_address = CRM_Utils_System::ipAddress();
    $contact_id = CRM_Core_Session::getLoggedInContactID();
    $user_agent = $_SERVER['HTTP_USER_AGENT'] ?? 'unknown';
    $fingerprint = NULL;

    // Support proxies
    if (Civi::settings()->get('loginsecurity_proxy')) {
      if (!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
        $ip_address = $_SERVER['HTTP_X_FORWARDED_FOR'];
      }
    }

    // Ignore drush running cron to avoid issues around CiviMail
    if ($ip_address == '127.0.0.1' && !$session_id && $user_agent == 'unknown') {
      return FALSE;
    }

    if (!$contact_id) {
      $apilog = Civi::settings()->get('loginsecurity_apilog');
      if ($apilog && !empty($_REQUEST['api_key'])) {
        $contact_id = CRM_Core_DAO::singleValueQuery('SELECT id FROM civicrm_contact WHERE api_key = %1', [
          1 => [$_REQUEST['api_key'], 'String'],
        ]);

        // We force this, because otherwise we will have a lot of records
        $session_id = 'REST API';
        $fingerprint = 'REST API';
      }
    }

    if (!$contact_id) {
      return FALSE;
    }

    // Cleanup the user agent
    $user_agent = str_replace('Mozilla/5.0 ', '', $user_agent);
    $user_agent = str_replace(' (KHTML, like Gecko)', '', $user_agent);
    $user_agent = str_replace('(', '', $user_agent);
    $user_agent = str_replace(')', '', $user_agent);

    $dao = CRM_Core_DAO::executeQuery('SELECT *, UNIX_TIMESTAMP(last_seen) as last_seen_ts
      FROM civicrm_login_security_device
      WHERE contact_id = %1 AND ip_address = %2 AND session_id = %3
      ORDER BY id ASC LIMIT 1', [
      1 => [$contact_id, 'Positive'],
      2 => [$ip_address, 'String'],
      3 => [$session_id, 'String'],
    ]);

    if ($dao->fetch()) {
      // Update the last_seen every 15 mins
      $now = time();
      $update_delay = 900;

      if ($now - $dao->last_seen_ts > $update_delay) {
        CRM_Core_DAO::executeQuery('UPDATE civicrm_login_security_device SET last_seen = NOW(), user_agent = %3 WHERE contact_id = %1 AND ip_address = %2', [
          1 => [$contact_id, 'Positive'],
          2 => [$ip_address, 'String'],
          3 => [$user_agent, 'String'],
        ]);

        return $contact_id;
      }

      // Known session and no need to update the last_seen
      return FALSE;
    }

    $api = \Civi\Api4\LoginSecurityDevice::create(FALSE)
      ->addValue('contact_id', $contact_id)
      ->addValue('ip_address', $ip_address)
      ->addValue('session_id', $session_id)
      ->addValue('user_agent', $user_agent);

    if ($fingerprint) {
      $api->addValue('fingerprint', $fingerprint);
    }

    $api->execute();

    $notify_users = Civi::settings()->get('loginsecurity_notify_users');
    $notify_admins = Civi::settings()->get('loginsecurity_notify_admins');
    $admin_emails = Civi::settings()->get('loginsecurity_notify_admin_emails');
    $is_new = false;

    // If we are going to send a notification, check if it is a new device
    if ($notify_users || $notify_admins) {
      $api4 = \Civi\Api4\LoginSecurityDevice::get(false)
        ->addSelect('COUNT(*) AS count')
        ->addWhere('contact_id', '=', $contact_id)
        ->addWhere('ip_address', '=', $ip_address);

      if ($fingerprint) {
        $api->addValue('fingerprint', $fingerprint);
      }

      $result = $api4->execute()->first();
      $is_new = $result['count'] == 1;
    }

    if ($notify_users == 'new' && $is_new || $notify_users == 'always') {
      $subject = $is_new ? E::ts('A new device or location has accessed your CiviCRM account') : E::ts('Your account has been used to access CiviCRM');

      self::sendNotification([
        'contact_id' => $contact_id,
        'ip_address' => $ip_address,
        'user_agent' => $user_agent,
        'user_notification_new' => $is_new,
        'loginsecurity_email_subject' => $subject,
      ]);
    }

    if (($notify_admins == 'new' && $is_new || $notify_admins == 'always') && $admin_emails) {
      $subject = $is_new ? E::ts('A new device or location has been used to access CiviCRM') : E::ts('An account has been used to access CiviCRM');
      $emails = explode(',', $admin_emails);

      foreach ($emails as $email) {
        $email = trim($email);

        self::sendNotification([
          'contact_id' => $contact_id,
          'ip_address' => $ip_address,
          'user_agent' => $user_agent,
          'admin_notification_new' => $is_new,
          'admin_email' => $email,
          'loginsecurity_email_subject' => $subject,
        ]);
      }
    }

    return $contact_id;
  }

  /**
   * Returns the GeoIP record information for a given IP.
   */
  public static function getGeoIP(String $ip) {
    // @todo Setting?
    $db_country = __DIR__ . '/../../../db/country.mmdb';
    $db_city = __DIR__ . '/../../../db/city.mmdb';

    // Avoid fatals if the files are not there
    if (empty(\Civi::$statics[__CLASS__]['autoload'])) {
      if (file_exists(__DIR__ . '/../../../vendor/autoload.php') && file_exists($db_country)) {
        require_once __DIR__ . '/../../../vendor/autoload.php';
        \Civi::$statics[__CLASS__]['autoload'] = 'success';
      }
      else {
        Civi::log()->warning('loginsecurity: the composer dependencies are missing. If you installed from git, please run "composer install" from the extension directory.');
        \Civi::$statics[__CLASS__]['autoload'] = 'failed';
      }
    }

    if (\Civi::$statics[__CLASS__]['autoload'] != 'success') {
      return null;
    }

    // Exceptions where the CiviCRM country does not match the MaxMind country
    // By using CiviCRM's country name, we can send the name to 'ts' for translation
    $country_civi_map = [
      'Russia' => 'Russian Federation',
    ];

    try {
      // @todo Make this more configurable, ex: if using a paid database?
      $reader = new \GeoIp2\Database\Reader($db_country);
      $record = $reader->country($ip);
      $country_name = $record->country->name;

      if (isset($country_civi_map[$country_name])) {
        $country_name = $country_civi_map[$country_name];
      }

      $country_name = ts($country_name, ['context' => 'country']);

      if (file_exists($db_city)) {
        $reader = new \GeoIp2\Database\Reader($db_city);
        $record = $reader->city($ip);
        $city_name = $record->city->name;

        if ($city_name) {
          if (in_array($record->country->name, ['United States', 'Canada'])) {
            return $city_name . ', ' . $record->subdivisions[0]->isoCode . ', ' . $country_name;
          }
          else {
            return $city_name . ', ' . $country_name;
          }
        }
      }

      return $country_name;
    }
    catch (Exception $e) {
      Civi::log()->warning('loginsecurity: Reader Exception: ' . $e->getMessage());
      return null;
    }

    return null;
  }

  /**
   *
   */
  public static function sendNotification($tplParams) {
    if (empty($tplParams['contact_id'])) {
      throw new Exception('Missing contact_id');
    }

    // Fetch the display_name of the user
    // Avoid confusion with display_name, because it might be an admin notification
    $contact = \Civi\Api4\Contact::get(false)
      ->addSelect('display_name')
      ->addSelect('email.email')
      ->setJoin([
        ['Email AS email', FALSE, NULL, ['email.is_primary', '=', 1]],
      ])
      ->addWhere('id', '=', $tplParams['contact_id'])
      ->execute()
      ->first();

    $tplParams['user_display_name'] = $contact['display_name'];

    if (!empty($tplParams['ip_address'])) {
      $tplParams['user_location'] = self::getGeoIP($tplParams['ip_address']);
    }

    // @todo Localize? or ISO-8601? Having the timezone is useful
    $tplParams['date_last_seen'] = date('Y-m-d H:i:s T');

    // About the site
    $tplParams['civicrm_url'] = CRM_Utils_System::url('civicrm', NULL, TRUE, NULL, FALSE, TRUE);
    $domainValues = CRM_Core_BAO_Domain::getNameAndEmail();

    $sendTemplateParams = [
      'groupName' => 'msg_tpl_workflow_uf',
      'valueName' => 'loginsecurity_access_notification',
      'tplParams' => $tplParams,
      'isTest' => 0,
      'from' => "$domainValues[0] <$domainValues[1]>",
    ];

    if (!empty($tplParams['admin_email'])) {
      $sendTemplateParams['toEmail'] = $tplParams['admin_email'];
    }
    else {
      // Assuming we are sending to the user
      $sendTemplateParams['toName'] = $contact['display_name'];
      $sendTemplateParams['toEmail'] = $contact['email.email'];
    }

    list($sent) = CRM_Core_BAO_MessageTemplate::sendTemplate($sendTemplateParams);
  }

}
